#!/usr/bin/perl -w

# Copyright (C) 2016, Teus Hagen, the Netherlands
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# If yuo have improvements please do not hesitate to email the author.

# $Id: showPM.pl,v 1.1 2017/03/28 16:53:55 teus Exp teus $
my $Version = '$Revision: 1.1 $, $Date: 2017/03/28 16:53:55 $.';
$Version =~ s/\$//g;
# Description:
# script will generate JS script Highchart graphs from pollutant data.
# The data is collected from the measurement stations and sensors tables
# in the database.
# TO DO:
#       allow the user to select different the levels for EU/WHO/LKI/AQI/AQI_LKI etc.
use constant {
                                        # Default DB credentials
        USER => 'someuser', PASSWD => 'somepass', HOST => 'localhost',
        DB        => 'luchtmetingen',   # default database

        TITLE     => 'Samen Meten aan Luchtkwaliteit (testfase fijnstofmetingen)', # chart
        SUBTITLE  => '',                                               # chart

        # following defaults can be altered on the command line with an option choice
        LOCATION  => 'Grubbenvorst',    # default location nr, sensor label
        REGION    => 'Horst aan de Maas',       # default region
        ID        => '13033927',        # default table of hourly measurements
        COORDINATES => '51.420698,6.135656', # default location coordinates
        PROJECT   => 'VW',              # default project id
        SENSORS   => 'Sensors',         # name of sensors desciption table
        REF       => 'HadM',            # most recent measurement station PLIM
        POLLUTANTS => '(pm|pm_10|pm_25)', # choice of pollutants to show in graph

        AQI       => 'EU',              # plotbands accoringly to EU/WHO,LKI,AQI
        MINHOURS  => 12,                # minimal nr hours to show a graph
        MINGRAPHS => 2,                 # minimal nr of graphs requirement
        PERIOD    => 60*60*24*7*3,      # total of period on chart 3 weeks
        WDIR      => '/webdata/luchtmetingen/',   # working directory
                                        # the website dir for chart page
        WEBDIR    => '/webdata/Drupal/cmsdata/BdP/files/luchtmetingen/',
        WWWGRP    => 'www-data',        # if output file name has full path name

        FALSE     => 0,
        TRUE      => 1,
};

use JSON;
use Time::Piece;
use POSIX qw(strftime);
use feature "state";
use Env qw(HOME DBUSER DBPASS DBHOST);  # for working from HOME dir, and DB credits
use feature "state";            # some routines keep own state variables
use POSIX qw(strftime);
use Time::Piece;                # need to parse time strings
use autodie;
use 5.010;
use LWP::UserAgent;             # for direct http access
use URI::Escape;
use Getopt::Mixed;              # need to parse command line arguments

my $debug = 0;   # on debug output full HTML page to output
my $verbose = 0; # verbosity level
my $myhost   = HOST;    # mysql server
$myhost      = $DBHOST if defined $DBHOST;
my $mydb     = DB;      # mysql database name
$mydb        = $DB if defined $DB;
my $myuser   = USER;    # mysql user
$myuser      = $DBUSER if defined $DBUSER;
my $mypass   = PASSWD;  # mysql password
if( $mypass =~ /somepass/ ){
    $mypass = '';
    $mypass = $DBPASS if defined $DBPASS;
}

use DBI;
my $mysql;

END {                   # on exit close connections
    if( $mysql and $mysql != 1 ){ $mysql->disconnect(); $mysql = 0; }
}

# make sure we work in a well known dir
my $WDir = WDIR;
if( -d $WDir ){
   chdir( $WDir) ;
} else {
   print STDERR "Cannot chdir to working dir $WDir\n";
   $WDir = './';
}

Getopt::Mixed::init(
        'H help>H '.
        'd:i debug>d '.
        'v verbose>v '.
        'q quiet>q '.
        'p=s pass>p '.
        'u=s user>u '.
        'h=s host>h '.
        'D=s database>D '.
        'l=s location>l '.
        'i=s ID>i '.
        'n nolast>n '.
        'P=s project>P '.
        'e=s pollutants>e '.
        'g=i graphs>g '.
        'O=s output>O ',
        'a:i aggregation>a '.
        't=s title>t '.
        's=s subtitle>s '.
        'r=s region>r '.
        'A=s aqi>A '.
        'AQI>A '.
        'index>A '.
        'w=s web>w '.
        'W=s wd>W '.
        ''
);
my $label = LOCATION;
my $identifier = ID;
my $coordinates = COORDINATES;
my $project = PROJECT;
my $title = TITLE;
my $subtitle = SUBTITLE;
my $mingraphs = MINGRAPHS;
my $region = REGION;
my $AQI = AQI;
my $webdir = WEBDIR;
my $wwwgrp = WWWGRP;
my $pollutants = POLLUTANTS;
my $output = '/dev/stdout';
my $aggregation = 12;

while( my($option, $value, $arg) = Getopt::Mixed::nextOption() ){
  OPTION: {
    $option eq 'd' and do { $verbose++; $debug++;  last OPTION; };
    $option eq 'v' and do { $verbose++;  last OPTION; };
    $option eq 'a' and do { $aggregation = 1;
            $aggregation = $value if defined $value;
            $aggregation = 1 if ($aggregation < 1);
            $aggregation = 12 if ($aggregation >= 12);
            last OPTION;
        };
    $option eq 'q' and do { $verbose = 0;  last OPTION; };
    $option eq 'h' and do { $myhost = $value; $myhost =~ s/\s//g; last OPTION; };
    $option eq 'D' and do { $mydb = $value; $mydb =~ s/\s//g; last OPTION; };
    $option eq 'e' and do { $pollutants = $value;
            $value =~ s/^/,/; $value =~ s/,(pm(_(10|25)?)|temp|luchtdruk|ws|[a-z]+[0-3]?)//g;
            if( $value ) {
                print STDERR "Unknown pollutant in pollutant choice: $pollutants.\n";
                $pollutants = POLLUTANTS;
            }
            $pollutants =~ s/,/|/g;
            $pollutants = "($pollutants)" if $pollutants !~ /^\(/;
            last OPTION;
        };
    $option eq 'u' and do { $myuser = $value; $myuser =~ s/\s//g; last OPTION; };
    $option eq 'p' and do { $mypass = $value; last OPTION; };
    $option eq 'P' and do { $project = $value; last OPTION; };
    $option eq 'g' and do { $mingraphs = $value; last OPTION; };
    $option eq 'l' and do { $label = $value; $label =~ s/\s//g; last OPTION; };
    $option eq 't' and do { $title = $value; last OPTION; };
    $option eq 's' and do { $subtitle = $value; last OPTION; };
    $option eq 'r' and do { $region = $value; last OPTION; };
    $option eq 'i' and do {
            $identifier = $value; $identifier =~ s/\s//g;
            last OPTION;
        };
    $option eq 'W' and do { $WDir = $value; chdir(WDir); last OPTION; };
    $option eq 'w' and do { $webdir = $value; last OPTION; };
    $option eq 'n' and do { $no_last = 1; last OPTION; };
    $option eq 'O' and do { $output = $value; last OPTION; };
    $option eq 'A' and do { $AQI = $value; last OPTION; };
    $option eq 'H' and do {
        my $ref = REF;
        print STDERR <<EOF ;
        $0 [options] [arg...] (default/no arguments: catch up to latest sensor data
            for the default location/identifier $identifier)
        $0 argument uses the name, optionaly search meetnetdata.rivm.nl site for
        the name and identier. Adds the info to the Sensors database. And
        downloads the data and/or inserts the data into the database 'luchtmetingen'.
        Default location or table to be used is ${project}_${identifier} (label_name ${project} sensor).
        Tooltip synchronisation: The start time of the graphs are shifted to the
        minute in the hour the first graph was started. Usually the low budget
        sensor graph.
        First try:
        $0 -d -v -O ./file_name ${project}_${identifier} $ref as first test.
        The -d for debug mode, -v be more versatile,
        ${project}_${identifier}, station $ref DB table names should be
        present in database $mydb.
        By default the graphs of PM2.5 (ie also the cheaper PM sensor) is made visible.
        The others can be switched on.

        Options:
 -O|--output    The output file with HTML/JS code. Default: $output
                If the output filename has no slash the file will be stored
                in the website directory (default: $webdir).
 -l|--location  The measurement location to use, default ($identifier) location $label
 -t|--title     The title in the chart, default '$title'.
 -s|--subtitle  The subtitle in the chart, default '$subtitle'.
 -r|--region    The region of measurements, default '$region'.
 -i|--ID        The measurement identifier to use, default $identifier
                The database table is defined by {project_ID}_identifier (${project}_$identifier)
 -d|--debug     Debugging on.
 -H|--help      This help or usage message.
 -v|--verbose   Increase verbosity, default: off.
 -q|--quiet     Be as quiet as possible, turns verbose off, default: off.

 -D|--database  The Database name to use (default: $mydb). Or env variable DB.
 -h|--host      The DB host for DB access (default: $myhost). Or env var. DBHOST.
 -u|--user      The DB user for DB access (default: $myuser). Or env. var. DBUSER.
 -p|--pass      The DB user password for DB access (default: X?x?X). Or env. var. DBPASS.

 -P|--project   The project name (default: VW). Is used as component in data DBtable name.
 -g|--graphs    The minimal amount of graphs in the chart (default: $mingraphs).
 -e|--pollutants The pollutants choice for the graphs shown (default: $pollutants)
                The pollutant identifier is the DB table column name: eg pm, o3, temp
 -A|--{AQI,aqi,index} {LKI|AQI|EU|none} colored bands in background for quality levels.
 -n|--nolast    Update DB with all values found and not from last value found.
                Default: update from last value found.
 -W|--wdir      The working directory with own modules.
 -w|--web       The website private page directory to store the page.
 -a|--aggregation [number] Switch 5 minutes to 1 hour aggregation off or pm count/hour
                count for eg 5 minutes avg counters. Default: 12 (aggregate to 1 hour)..
                Minimum aggregation value is 1 (every 5 minutes), max is 12 (1 hour).

$Version
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation.
The script depends om mysql, yui-compressor,
and perl modules (JSON,Posix,,Time,autodie,LWP,URI and DBI). 
See the script constant declaration for the default configuration values.
EOF
        exit(0);
        };
   };
}
Getopt::Mixed::cleanup();
$verbose = 3 if $debug;

die "FATAL ERROR: no database user/password defined, e.g. define use environment variable DBUSER and/or DBPASS\n" if length($mypass) <= 0;


sub Check_DB { # called only once from first call to query routine
    state $checkedOnce = 1;
    if( $mysql ){ return 1; }
    return 0 if length( $mydb ) < 1;
    return 0 if length( $myhost ) < 1;
    return 0 if length( $myuser ) < 1;
    return 0 if length( $mypass ) < 1;
    $mysql = DBI->connect("DBI:mysql:".$mydb.":".$myhost,$myuser,$mypass);
    if( not $mysql ){
        if( $checkedOnce > 0 ){
            print STDERR "ERROR Cannot open mysql database ".$mydb." on host ".$myhost.", with user ".$myuser.", error: ". DBI->errstr."\n" ;
            exit 1;
        }
    }
    return 1;
}

my %DB_cols = ( );      # sensors names per table
# obtain all sensors for a station in the DB table
sub Get_Sensors {
    my $tbl = shift; return 0 if not $tbl;
    Check_Tbl($tbl) if not $DB_cols{$tbl};
    if( not defined $DB_cols{$tbl}{"id"} ){
        # only once at the start we build a column existance cache
        my $qr = query($tbl, "DESCRIBE $tbl;");
        if( (not $qr) || ($#{$qr} < 0) ){
            print STDERR "ERROR Cannot obtain table $tbl description\n";
            return 0;
        }
        # mysql counts cells from 0!
        $DB_nr_cols{$tbl} = 0 if not defined $DB_nr_cols{$tbl};
        for( my $index = 0; $index <= $#{$qr} ; $index++ ){
            $DB_cols{$tbl}{ $qr->[$index] } = 1;
            next if $qr->[$index] =~ /^(id|datum|.*_(valid|ppb))$/ ;
            $DB_nr_cols{$tbl}++;
            $DB_cols{$tbl}{$qr->[$index]} = 1;
        }
    }
    return 1;
}

# check if sensor is available in table if not add the column
sub Check_Sensor {
    my $tbl = shift;
    my $col = lc(shift);
    my $comment = shift;
    return 0 if not Get_Sensors($tbl);
    return 1 if defined $DB_cols{$tbl}{$col};
    return 0;
}

# one row of values from one column out of the DB table
sub query {
    return 0 if Check_DB == 0;
    my $tbl = shift; Check_Tbl($tbl) if $tbl;
    my $q = shift;
    if( (not $mysql) or ($mysql == 1) or (not $q) ){ return undef; }
    print STDERR "MYSQL: $q\n" if $debug > 1;
    my $sth = $mysql->prepare($q); $sth->execute();
    #  ref to 2 dimensional refs to array string values
    if( $q =~ /^\s*(show|describe|select)/i ){
        my $r = $sth->fetchall_arrayref();
        print STDERR "ERROR mysql query: $q with error:\n" . DBI->errstr . "\n"
            if ( DBI->errstr &&  $#{$r} < 0 );
        return undef unless $r->[0];
        my @rts = ();
        for( my $i=0; $i <= $#{$r}; $i++ ){
            $rts[$i] = $r->[$i][0] if $#{$r->[$i]} >= 0;
        }
        if( not @rts ){
            # print STDERR "$WARNING mysql query: $q with empty array:\n" . DBI->errstr . "\n";
        }
        $sth->finish(); return \@rts;
    } else { return (DBI->errstr ? undef : 1); }
}

# matrix of values
sub long_query {
    return 0 if Check_DB == 0;
    my $tbl = shift; Check_Tbl($tbl) if $tbl;
    my $q = shift;
    if( (not $mysql) or ($mysql == 1) or (not $q) ){ return undef; }
    print STDERR "MYSQL: $q\n" if $debug > 1;
    my $sth = $mysql->prepare($q); $sth->execute();
    #  ref to 2 dimensional refs to array string values
    if( $q =~ /^\s*(select)/i ){
        my $r = $sth->fetchall_arrayref();
        return undef unless $r->[0];
        if( $#{$r} < 0 ){
            print STDERR "ERROR mysql query: $q with error:\n" . DBI->errstr . "\n";
        }
        $sth->finish(); return $r;
    } else { return undef; }
}

# collect from database the table column names
# if table does not exists create it
sub Check_Tbl {
    my $tbl = shift;
    return 0 if not $tbl;
    return 1 if defined $DB_cols{$tbl};
    my $qry = query('',"SHOW TABLES");
    for( my $i = 0; $i <= $#{$qry}; $i++ ){
        next if $qry->[$i] !~ /^$tbl$/;
        $DB_cols{$qry->[$i]} = {} if not $DB_cols{$qry->[$i]}; # the table already exists
        Get_Sensors($qry->[$i]);        # mark sensors present in the table
    }
    return 1 if defined $DB_cols{$tbl};
    return 0;
}

# get last date of entered values of the sensor
sub Last_date {
    return 0 if (defined $no_last) && (not $no_last);
    my $table = shift;
    my $rslt = query($table, "SELECT UNIX_TIMESTAMP(datum) FROM $table ORDER BY datum DESC LIMIT 1");
    return 0 if (not defined $rslt) || ($#{$rslt} < 0);
    return $rslt->[0];
}

# collect info from database SENSORS table. Return hash table with info
sub Get_Info {
    my $id = shift; return undef if (not defined $id) || (not $id);
    my $qry; my %info; my $indx = '';
    my $sensors = SENSORS; Check_Tbl($sensors);
    foreach my $i ('serial','label') {
        $qry = query($sensors,"SELECT count(*) FROM $sensors WHERE  $i = '$id'");
        if( $qry->[0] ) {
            $indx = $i; $info{$i} = $id; last;
        }
    }
    return %info if not $indx;
    $qry = long_query($sensors,"SELECT
            serial,label,coordinates,UNIX_TIMESTAMP(first),street,village,description
            FROM $sensors WHERE $indx = '$id' ORDER BY datum DESC");
    return undef if ( ($#{$qry} < 0) || (not $qry->[0]) || ($#{$qry->[0]} < 0) ) ? 1:0;
    my @in = @{$qry->[0]};
    for( my $i = 0; $i <= $#in; $i++){ $in[$i] = '' if not defined $in[$i]; }
    $info{serial}=$in[0];
    $info{label}=$in[1];
    $info{coordinates}=$in[2];
    $info{first}=$in[3] if defined $in[3];
    $info{street}=$in[4] if defined $in[4];
    $info{village}=$in[5] if defined $in[5];
    $info{location}=$in[4] if defined $in[4];
    $info{location} = $info{location} . ', '.$info{village}
        if defined $in[5];
    $info{location} = $info{label} if not defined $info{location};
    $info{description} = $in[6];
    return %info;
}

# timestamp of last value from sensor
sub Get_last {
    my $Ref = shift; $Ref = REF if not defined $Ref;
    my $thisLast = 0; my $pol = '';
    # check the available pollutants from the pollutant choice
    my @pols; $pols[0] = $pollutants; $pols[0] =~ s/[\(\)]//;
    @pols = split(/\|/,$pols[0]);
    foreach (@pols) {
        $pol = $_; last if  Check_Sensor($Ref,$_);
        $pol = '';
    }
    if( not $pol ) {
        print STDERR "Could not find PM type in $Ref table\n";
        return 0;
    }
    my $qry = query($Ref,"SELECT UNIX_TIMESTAMP(datum) FROM $Ref WHERE ${pol}_valid ORDER BY datum DESC LIMIT 1");
    if( ($#{$qry} < 0) || ($qry->[0] < 1000) ) {
        print STDERR "Cannot find last measurement of $pol for $Ref\n" ;
        return 0;
    }
    return $qry->[0];
}

# combine an array to one value. Missing values are filled with avg over the array
sub aggr2one {
    my $window = shift; my $cnt = 0; my $val = 0;
    for( my $i= 0; $i <= $#{$window}; $i++ ) {
        next if (not defined $window->[$i]) || (not $window->[$i]);
        $cnt++; $val += $window->[$i];
    }
    return 'null' if not $cnt;
    $val += int((($val/$cnt)*($#{$window}+1-$cnt))*10+0.5)/10;
    return 'null' if not $val;
    return $val;
}
        
# aggregate with window $agg values in an array
sub aggregate {
    my ($values,$agg) = @_; my @rslt;
    for( my $i = 0; $i <= $#{$values}; $i += $agg) {
        if( $agg > 1 ) {
            my @window;
            for( my $j = $i; $j < ($i+$agg); $j++) {
                push @window, $values->[$j] if defined $values->[$j];
            }
            $#window = $agg-1;
            push @rslt,aggr2one(\@window);
        } else {
            push @rslt, ($values->[$i]? $values->[$i] : 'null');
        }
    }
    return join ',',@rslt;
}

# extract data from DB for a pollutant and serialize it into json data struct
# return ref to hash table with name, location, pollutant, string as json data
sub Get_data {
    my ($tbl,$pol,$last,$secs,$period) = @_; my %rslt;
    $period = PERIOD if not defined $period;
    return \%rslt if not defined $DB_cols{$tbl}{$pol}; 
    my $cnt = 0;
    my $data = query($tbl,"SELECT CONCAT(UNIX_TIMESTAMP(datum),'=',$pol) FROM $tbl
                WHERE (UNIX_TIMESTAMP(datum) >= ($last-$period)) AND 
                    (UNIX_TIMESTAMP(datum) <= $last) AND
                    NOT ISNULL($pol) AND ${pol}_valid
                    ORDER BY datum");
    return \%rslt if $#{$data} < MINHOURS*(60*60/$secs); # MINHOURS hours minimal
    $rslt{pol}=$pol; $rslt{table}=$tbl; $rslt{data} = '['; $rslt{first} = 0;
    my $time = 0; my $value = 0; my @values;
    my $last_time = 0; # if from real start: $last-$period
    for( my $i = 0; $i <= $#{$data}; $i++ ){
        $value = $time = $data->[$i];
        $time =~ s/=.*//; $value =~ s/.*=//; 
        $last_time = $rslt{first} = $time if not $last_time;
        while( ($last_time+$secs) < $time ) {
            push @values,0; $last_time += $secs;
        }
        $last_time = $time;
        push @values,$value; $cnt++;
    }
    my $end = $time;
    # while( $time < $last ) { push @values,0; $time += $secs; }
    $cur_agg = ($secs >= 60*60? 1: $aggregation);
    $rslt{last} = $time; $rslt{unit} = $secs*$cur_agg;
    $rslt{data} .= aggregate(\@values,$cur_agg) . "]";
    $rslt{data} = undef if $cnt < ($cur_agg*12);  # at least for 12 hours values
    my $cnt2 = () = $rslt{data} =~ /,/g; $cnt2++;
    printf STDERR ("WARNING: for table %s, period %d days ending at %s: not enough values found: %d values, expected %d values.\n",
            $tbl, int(($period/(24*60*60))+0.5),
            strftime("%Y-%m-%d %H:%M",localtime($last)),
            $cnt, int(($period/$secs)+0.5) )
        if not defined $rslt{data};
    printf STDERR ("For table %s, sensor %s: found %d values,\n\tstarts at %s, ends at %s,\n\tgot %d values, expected %d%s values.\n",
            $tbl, $pol, $cnt, 
            strftime("%Y-%m-%d %H:%M",localtime($rslt{first})),
            strftime("%Y-%m-%d %H:%M",localtime($end)),
            $cnt2, int(($period/$secs/($cur_agg> 1?$aggregation:1))+1.5),
            ($cur_agg> 1? " (aggregration: $aggregation)":'') )
        if $verbose > 1;
    return \%rslt;
}

# collect the data for a serie of stations (DB table names)
# return a ref to array of hashes: table name, location name, pm name
sub Collect_data {
    my @stations = @_; my $last = Get_last(REF); my %info;
    my @data; my $first = time;
    if( not $last ) {
        print STDERR "Cannot find last measurement date/time\n";
        return 0;
    }
    for( my $i = 0; $i <= $#stations; $i++ ) {
        if( $stations[$i] =~ /^(VW_[0-9]+|[A-Z][a-z]+)$/ ) {
            %info = Get_Info($stations[$i]);
            print STDERR "Cannot find station $stations[$i] in database.\n"
                if not defined $info{serial};
            next if not defined $info{serial};
            my $tbl = 'VW_'.$info{serial};
            next if not Check_Tbl($tbl);
            Get_Sensors($tbl);
            foreach my $S (keys %{$DB_cols{$tbl}}) {
                next if $S =~ /(^id|^datum|_valid)/;
                # may add here more sensors
                next if $S !~ /$pollutants/i; # only pollutants choice
                my $D = Get_data($tbl,$S,$last,5*60);
                next if not defined $D->{data};
                my $c = () = $D->{data} =~ /,/g;
                next if $c < 12;            # minimal 12 hours of values
                $first = $D->{first} if $D->{first} < $first;
                $last = $D->{last} if $D->{last} > $last;
                $D->{location} = $info{location};
                $D->{label} = $info{label}; # mark for low cost sensor
                $D->{serial} = $info{serial}; # provide serial nr fot Sensors lookup
                $D->{sense} = $S;
                $D->{organisation} = 'RIVM';
                $D->{description} = $info{description};
                $D->{table} = "VW_$info{serial}";
                if( $#data < 0 ) { # always first in array
                    push @data, $D;
                } else {
                    push @data, $data[0];
                    $data[0] = $D;
                }
            }
        } else {
            if( not Check_Tbl($stations[$i]) ){
                print STDERR "Cannot find station $stations[$i] in database.\n";
                next;
            }
            Get_Sensors($stations[$i]);
            my $qry = query("stations","SELECT CONCAT(name,';',municipality,';',organisation) 
                        FROM stations WHERE stations.table = '$stations[$i]' LIMIT 1");
            my @info; @info = split /;/,$qry->[0] if $#{$qry} >= 0;
            $info[0] =~ s/^[A-Za-z]+\s//;  # shorten the street string a bit
            foreach my $S (keys %{$DB_cols{$stations[$i]}}) {
                next if $S =~ /(^id|^datum|_(valid|color|pp))/;
                # may add here more sensors
                next if $S !~ /$pollutants/; # only pollutants of choice
                my $D = Get_data($stations[$i],$S,$last,60*60);
                next if not defined $D->{data};
                my $c = () = $D->{data} =~ /,/g;
                next if $c < 12;            # minimal 12 hours of values
                $D->{sense} = $_;
                $first = $D->{first} if $D->{first} < $first;
                $last = $D->{last} if $D->{last} > $last;
                $D->{location} = $info[0];
                $D->{location} .= ", $info[1]" if defined $info[1];
                $D->{organisation} = $info[2] if defined $info[2];
                $D->{table} = $stations[$i];
                push @data, $D;
            }
        }
    }
    printf STDERR ("Collected chart data from %s upto %s\n",
              strftime("%Y-%m-%d %H:%M",localtime($first)),
              strftime("%Y-%m-%d %H:%M",localtime($last)) )
          if $verbose;
    $#data = -1 if ($#data+1) < $mingraphs;  # at least MINGRAPHS graphs
    push @data, $first, $last if $#data > 0; # end of array has first and last date
    return \@data;
}
            
# generate plotBands none,EU-WHO or LKI
# boundaries are for PM2.5  [from,to,quality mark]
my %bands = (
    EU => [
        [0,20,'EU norm, PM\u2081\u2080 fase 2'],
        [20,25,'WHO norm, PM\u2082.\u2085'],
        [25,40,'EU norm, PM\u2081\u2080 jaargem.'],
        [40,50,'EU norm, PM\u2081\u2080 daggemiddelde'],
        [50,200,''],
        ],
    LKI => [
        [0,20,'LKI Index PM\u2082.\u2085 goed'],
        [20,50,'LKI Index PM\u2082.\u2085 matig'],
        [50,89,'LKI Index PM\u2082.\u2085 ongezond'],
        [89,200,'LKI Index PM\u2082.\u2085 slecht'],
        ],
    AQI => [
        [0,12,'AQI Index PM\u2082.\u2085 goed'],
        [12,36,'AQI Index PM\u2082.\u2085 matig'],
        [36,56,'AQI Index PM\u2082.\u2085 opgepast'],
        [56,150,'AQI Index PM\u2082.\u2085 ongezond'],
        [150,200,'AQI Index PM\u2082.\u2085 gevaarlijk'],
        ],
    AQI_LKI => [
        [0,12,'AQI Index PM\u2082.\u2085 goed'],
        [12,20,'AQI Index PM\u2082.\u2085 matig'],
        [20,36,'LKI Index PM\u2082.\u2085 matig'],
        [36,50,'AQI Index PM\u2082.\u2085 opgepast'],
        [50,56,'LKI Index PM\u2082.\u2085 ongezond'],
        [56,89,'AQI Index PM\u2082.\u2085 ongezond'],
        [89,150,'LKI Index PM\u2082.\u2085 slecht'],
        [150,200,'AQI Index PM\u2082.\u2085 gevaarlijk'],
        ],
);
        
sub plotBands {
    my $NRM = shift;
    return '' if (not $NRM) || ($NRM =~ /none/);
    print STDERR "Unknown quality level identifier $NRM\n" if not defined $bands{$NRM};
    return '' if not defined $bands{$NRM};
    my @colors = (
        # maybe we shoud use the colors as defined by the ref doc, see AQI.pl
        ['0, 32, 197, 0.05','0, 32, 197, 0.45'],
        ['244, 230, 69, 0.15','175, 158, 44, 1'], # yellow text needs more contrast
        ['254, 118, 38, 0.15','254, 118, 38, 0.45'],
        ['220, 6, 16, 0.15','220, 6, 16, 0.45'],
        ['162,23,148,0.15','162,23,148,0.45'],
        ['162,23,148,0.15','162,23,148,0.45'],
        ['162,23,148,0.15','162,23,148,0.45'],
        ['162,23,148,0.15','162,23,148,0.45'],
    );
    my $template = "
        {color:'rgba(COLB)',from:FROM,to:TO,
            label:{
                text:'TEXT',
                rotation:0,
                align: 'center',
                verticalAlign: 'top',
                style:{
                    color:'rgba(COLT)',
                    fontWeight:'bold',
                    fontSize: '80%',
                },
                y: 10,
            }
        },
        ";
    my @levels = @{$bands{$NRM}}; my $rslt = '';
    for( my $i = 0; $i <= $#levels; $i++ ) {
        my $strg = $template; my ($from,$to,$text) = @{$levels[$i]};
        last if $i > $#colors;
        my $colb = $colors[$i][0]; my $colt = $colors[$i][1];
        $to -= 0.30; # keep a white line
        $strg =~ s/FROM/$from/m; $strg =~ s/TO/$to/m;
        $strg =~ s/COLB/$colb/m; $strg =~ s/COLT/$colt/m;
        $strg =~ s/TEXT/$text/m;
        $rslt .= $strg;
    }
    return $rslt;
}

# compress JS script string
sub JScompress {
    my $string = shift; my $rslt = '';
    return ${$string} if $debug;
    return '' if not ${$string};
    my $JSin;
    open $JSin,  ">/var/tmp/VW_IN.js" || return ${$string};
    print $JSin ${$string}; close $JSin;
    local $/;
    if( not -x "/usr/bin/yui-compressor" ) {
        print STDERR "WARNING yui-compressor is not installed. No JS compression.\n";
        return ${$string};
    }
    open $JSin, "/usr/bin/yui-compressor --type js /var/tmp/VW_IN.js 2>/dev/null |" || return ${$string};
    
    $rslt = <$JSin>;
    if ( not $rslt ) {
        print STDERR "WARNING yui-compressor failed. JS data saved in /var/tmp/VW_IN.js\n";
        print STDERR "No JS compression is done.\n";
        return ${$string};
    }
    close $JSin;
    system("/bin/rm -f /var/tmp/VW_IN.js");
    return $rslt;
}

# get a row from DB Sensors with first date/time, location in lifetime sensor
# ordered by first date/time entry in table of this sensor
sub Get_Places{
    my ($stations,$serial,$label,$pollutant) = @_;
    $pollutant = " ($pollutant)" if (defined $pollutant) && $pollutant;
    my $rslt = "0,'?$pollutant'"; my @recs; my $qry;
    Check_Tbl($stations);
    if( $stations eq SENSORS ) {
        return $rslt if not $serial;
        $serial = "'$serial' AND label = '$label'"
            if (defined $label) && $label;
        my $records = 'UNIX_TIMESTAMP(first)*1000,description,street,village,municipality';
        $qry = long_query($stations,"SELECT $records FROM $stations
            WHERE serial = $serial ORDER BY first");
        $records =~ s/^[^,]*\(([a-z]+)\)[^,]*,/$1,/;
        @recs = split /,\s*/, $records;
    } else {
        my $records = 'first,name,municipality'; # one may use geolocation field
        $qry = long_query($stations,"SELECT $records FROM $stations
            WHERE station.table = '$serial'");
        @recs = split /,\s*/, $records;
    }
    return $rslt if $#{$qry} < 0;
    for( my $i = 0; $i <= $#{$qry}; $i++ ){
        next if $#{$qry->[$i]} < 0; my $desc = '?';
        for( my $j = 1; $j <= $#recs; $j++ ){
            if( (defined $qry->[$i][$j]) && $qry->[$i][$j] ) {
                $rslt .= ',' if $rslt;
                $rslt .= "$qry->[$i][0], '$qry->[$i][$j]$pollutant'";
                last;
            }
        }
    }
    return $rslt;
}
    
my $OUT;
sub MyPrint {
    my $inscript = shift; my $strg = shift;
    state $JS = '';
    return 0 if not $strg;
    if ( $debug ) {
        print $OUT $strg;
    } elsif ( $inscript ) {
        $JS .= $strg;
    } else {
        if ( $JS ) {
            print $OUT JScompress(\$JS);
            $JS = '';
        }
        print $OUT $strg;
    }
}

# main routine, generate the HTML data for some stations from a template
sub Generate {
    my @stations = @_; my $data;
    my $IN;
    # strings to be used in generated HTML code
    my $Mlbls = ''; my $Slbls = ''; my $first = 0; my $last = 0;
    $data = Collect_data(@stations);
    if( $#{$data} <= 3 ) {
        printf STDERR ("Unable to find chart data for %s stations\n",
            join(', ', @stations));
        return 0;
    }
    $last = strftime("%a %e %b %Y %H:%M", localtime(pop @{$data}));
    $first = strftime("%a %e %b %Y", localtime(pop @{$data}));
    printf STDERR ("Found data for %d sensors.\n", $#{$data}+1)
        if $verbose;
    my $prev = 'X';
    my $timeshift = 0;  # we try to start all plot intervals on same minute
    # so the tooltip values will be shown in parallel
    # the stations have avg values per hour. It is unclear from the documentation
    # if the given value is of the previous hour, or that hour.
    # The database thinks the time is half of the hour: 30 minute boundary.
    # the shift is done on the last start time provided. Usualy a station.
    for( my $i = 0; $i <= $#{$data}; $i++ ){    # collect all locations
        $timeshift = $data->[0]{first}%3600 if $i == 0;
        $data->[$i]{first} -= ($data->[$i]{first}%3600) - $timeshift; # shift graph
        my $loc = $data->[$i]{location}; # $loc =~ s/^(.*)\s*,\s*(.*)$/$1 \($2\)/;
        my $S = $data->[$i]{pol}; $S = uc($S) if $S !~ /(ws|rv|temp|luchtdruk)/;
        $data->[$i]{sense} = $data->[$i]{pol};
        $S =~ s/PM_/PM/; $S =~ s/PM25/PM2.5/;
        $S =~ s/([A-Z])([0-9\.]+)/$1<span style="font-size:80\%">$2<\/span>/g;
        $data->[$i]{pol}=$S;
        if( $loc ne $prev ){   # create a text list in Dutch language
            if( defined $data->[$i]{label} ){
                if( $Slbls ){ $Slbls =~ s/ en /, /; $Slbls .= ' en '; }
                if( (not defined $data->[$i]{description}) || (not $data->[$i]{description}) ) {
                    $Slbls .= $loc;
                } else { $Slbls .= $data->[$i]{description} ; }
                $Slbls .= " ($data->[$i]{organisation})" if defined $data->[$i]{organisation}; 
            } else {
                if ( $Mlbls ) { $Mlbls =~ s/ en /, /; $Mlbls .= ' en '; }
                if( (not defined $data->[$i]{description}) || (not $data->[$i]{description}) ) {
                    $Mlbls .= $loc;
                } else { $Mlbls .= $data->[$i]{description} ; }
                $Mlbls .= " ($data->[$i]{organisation})" if defined $data->[$i]{organisation}; 
            }
        }
        $prev = $loc;
    }

    if( $output !~ /^\/dev\// ) {
        $output =~ s/\.html//i;
        $output .= '.html' if $debug;
        $output = $webdir . $output if $output !~ /\//;
    }
    if( $debug ) {
        $RSLT=sprintf("Debugmodus: HTML page ready output on %s, working dir %s\n",$output,WDIR);
    } else {
        $output .= '.html' if $output !~ /^\/dev\//;
        $RSLT=sprintf("File=%s with compressed JS script. Working directory: %s\n",$output,WDIR);
    }
    open $OUT, ">$output" || die ("Cannot open ${output} for generated HTML code\n");

    my $indoc = 0; my $skip = 0; my $inscript = 0;
    while( TRUE ){            # parse template file
        INPUT:
        if ( $IN ) { $_ = <$IN> ; }
        else { $_ = <main::DATA> }
        last if not $_;
        print STDERR $_ if $debug > 2;
        if( /<body/ ){ $indoc = 1;  next if not $debug; }
        if( /<\/body/ ){ $indoc = 0;  next if not $debug; }
        next if (not $debug) && (not $indoc);
        if( (/<script\s/) && $indoc ){
            MyPrint($inscript,$_); $inscript = 1; next;
        } 
        if( /<\/script>/ ){ $inscript = 0; };
        next if (not $indoc) && (not $debug);
        # output switcher: debug, dom ready function part and script global part
        if( /^(\/\/|<!--\s+)START\s+(DOM|GLOB)/ ) {
            my $sw = $2;
            if( $sw =~ /GLOB/ ){        # global var definitions
                # TO DO include organisations of sensors
                MyPrint($inscript,sprintf("var BdPtitle = '%s';\n", $title));
                MyPrint($inscript,sprintf("var BdPsubtitle = '%s';\n",
                    $subtitle ? $subtitle: 'regio '.$region));
                for( my $i = 0; $i <= $#{$data}; $i++ ){
                    MyPrint($inscript,sprintf("var BdPstart%d = %d*1000;\n", $i, $data->[$i]{first}));
                    MyPrint($inscript,sprintf("var BdPunit%d = %d*1000;\n",$i, $data->[$i]{unit}));
                    MyPrint($inscript,sprintf("var BdPdata%d = %s;\n",$i, $data->[$i]{data}));
                    if( defined $data->[$i]{label} ) {
                        $data->[$i]{places} = Get_Places(SENSORS,$data->[$i]{serial},$data->[$i]{label},$data->[$i]{pol});
                        if( $data->[$i]{places} =~ /,.*,/ ) {
                            MyPrint($inscript,sprintf("var BdPtitle%d = [%s];\n",
                                $i,$data->[$i]{places}) );
                        } else { # should not happen
                            my $choice = 'label';
                            foreach ('label','street','description') {
                                $choice = $_ if (defined $data->[$i]{$_}) && $data->[$i]{$_};
                            }
                            MyPrint($inscript,sprintf("var BdPtitle%d = [0,'%s (%s)'];\n",
                                $i, $data->[$i]{$choice}, $data->[$i]{pol}));
                        }
                    } else { # change this in a call to Get_Places('stations'...)
                        my $loc = $data->[$i]{location};
                        $loc =~ s/,\s[A-Za-z\s]+//;
                        MyPrint($inscript,sprintf("var BdPtitle%d = [0,'%s (%s)'];\n",
                            $i, $loc,$data->[$i]{pol}));
                    }
                }
            }
            $skip++ if $sw =~ /GLOB/; next;
        } elsif ( /^(\/\/|<!--\s+)END\s+(GLOB|DOM)/ ) {
            $skip = 0;
            next;
        }
        if( /^(\/\/\s*|<!--\s+)START\s+([a-zA-Z0-9]*).*/ ){ # insert new values
            my $type = $2; $skip++;
            if( $type =~ /regio/ ){              # for the region text
                MyPrint($inscript, $region . "\n");
            } elsif( $type =~ /stations/ ){      # all stations text
                MyPrint($inscript, "$Mlbls\n");
            } elsif( $type =~ /locations/ ){     # all locations of sensors text
                MyPrint($inscript, "$Slbls\n");
            } elsif( $type =~ /updated/ ){       # last update of chart
                MyPrint($inscript, strftime("%Y-%m-%d %H:%M\n",localtime(time)));
            } elsif( $type =~ /first/ ){         # first date in chart
                MyPrint($inscript, $first . "\n");
            } elsif( $type =~ /last/ ){          # last date in chart
                MyPrint($inscript, $last . "\n");
            } elsif( $type =~ /regio/ ){         # region text
                MyPrint($inscript, $region."\n");
            } elsif( $type =~ /straat/ ){        # street name text
                MyPrint($inscript, 'straat naam'."\n");
            } elsif( $type =~ /AQI/ ){           # AQI text lines
                # next should be compiled from %bands
                my %AQItxt = (
                    LKI => 'luchtkwaliteit LKI (RIVM) Index',
                    EU  => 'EU en WHO normen',
                    AQI => 'internationale luchtkwaliteit AQI Index',
                    # complete but complex overview and differences,
                    # discouraged to be shown
                    AQI_LKI => 'AQI en LKI luchtkwaliteits Index niveaux',
                );
                next if (not defined $AQItxt{$AQI}) || (not defined $bands{$AQI});
                my $levels = '';
                for(my $i = 1; $i <= $#{$bands{$AQI}}; $i++) {
                    # compile a list of levels
                    $level =~ s/ en /, / if $level;
                    $level .= ' en ' if $level;
                    $level .= $bands{$AQI}[$i][0];
                }
                $AQItxt{$AQI} .= ' grensen (resp. LEVEL';
                $AQItxt{$AQI} =~ s/LEVEL/$level/;
                $AQItxt{$AQI} .= ' &micro;g/m&sup3;)';
                MyPrint($inscript, $AQItxt{$AQI}."\n");
            } elsif( $type =~ /plotBands/ ){     # plotbands background levels
                MyPrint($inscript, plotBands($AQI));
            } elsif( $type =~ /series/ ){        # series with ref to data sructure
                for( my $i = 0; $i <= $#{$data}; $i++ ){
                    my $ugm3 = '\u00B5g/m\u00B3';
                    my $nr = '\u2248 pm/uur';
                    my $visible = 1;
                    $visible = 0 if not defined $data->[$i]{label};
                    # PM2.5 pollutants are all visible
                    $visible = 1 if (defined $data->[$i]{sense}) && ($data->[$i]{sense} =~ /pm_25/);
                    MyPrint($inscript,sprintf("
            { type: 'spline',
              pointStart: BdPstart%d,
              pointInterval: BdPunit%d,
              data: BdPdata%d,
              lineWidth: 1+%d,
              name: BdPtitle%d[BdPtitle%d.length-1],
              visible: %s,
              zIndex: 2,
              tooltip: {
                pointFormatter: function() {
                    var strg = 'none'; var i;
                    for( i = 0; i < BdPtitle%d.length; i += 2 ) {
                        if( this.x >= BdPtitle%d[i] ) { strg = BdPtitle%d[i+1]; }
                    }
                    return '<span style=\"color:' + this.color + '\">\\u25CF</span> ' + strg+': <b>' + this.y + ' %s</b><br/>';
                },
              },
              yAxis: %d,
              pointPlacement: 'between',
              marker:{
                radius: 1+%d
              }
            },\n",
                    $i,$i,$i, (defined $data->[$i]{label}?1:0),
                    $i,$i,
                    ($visible?'true':'false'),
                    $i,$i,$i, (defined $data->[$i]{label}?$nr:$ugm3),
                    (defined $data->[$i]{label}?0:1),
                    (defined $data->[$i]{label}?1:0) ) );
                } 
            }
            while( TRUE ){                         # skip rest
                if( $IN ){ $_ = <$IN> ; }
                else { $_ = <main::DATA> }
                if( not $_ ) { $skip = 0; last; }
                print STDERR $_ if $debug > 2;
                if( /^(\/\/\s*|<!--\s+)END/ ){
                    $skip = 0;
                    if( $IN ){ $_ = <$IN> ; }
                    else { $_ = <main::DATA> }
                    last;
                }
            }
        }
        if ( /VUURWERK/ ) {     # div identifier
            s/VUURWERK/SENSORS/g;
        }
        s/\/\/\s.*// if not $debug;
        s/<!--\s.*\s-->// if not $debug;
        next if /^\s*$/;
        MyPrint($inscript, $_) if not $skip;
    }
    system("/bin/chgrp $wwwgrp $output") if $output =~ /^\//;
    chmod 0644, $output;
    print STDERR $RSLT if $verbose;
    return 0;
}

die "Cannot access database $mydb. Correct user/password?\n" if not Check_DB();

if( $#ARGV >= 1 ) {     # at least 2 sensor locations
    Generate(@ARGV);
} else {                # default
    Generate(LOCATION,REF);
}

__DATA__
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"><html><head>
            <script src='https://ajax.googleapis.com/ajax/libs/jquery/1.12.4/jquery.min.js'></script>
            <script src='http://code.highcharts.com/stock/highstock.js'></script>
        
<script src='http://code.highcharts.com/highcharts-more.js'></script>
</head><body>
<script type="text/javascript">
//START GLOB data defs
//END GLOB data
$(function () { // on DOM ready
//START DOM
    Highcharts.setOptions({
        lang: {
                months: ['januari','februari','maart','april','mei','juni','juli','augustus','september','october','november','december'],
                shortMonths: ['jan','feb','mrt','apr','mei','jun','jul','aug','sep','oct','nov','dec'],
                weekdays: ['zondag','maandag','dinsdag','woensdag','donderdag','vrijdag','zaterdag'],
                shortWeekdays: ['zo','ma','di','wo','do','vr','za'],
                rangeSelectorTo: 'tot',
                rangeSelectorFrom: 'van',
            }
        }
    );
    $('#VUURWERK').highcharts('StockChart',{
            rangeSelector: {
            selected: 1,
            height: 40,
            buttonPosition: { y: 2 },
            buttons: [
                { type: 'day', count: 1, text: 'dag',
                 // dataGrouping: { forced: true, units:[['day',[1]]]}
                },
                { type: 'day', count: 3, text: '3 dagen',
                 // dataGrouping: { forced: true, units:[['day',[1]]]}
                },
                { type: 'week', count: 1, text: 'week',
                 // dataGrouping: { forced: true, units:[['day',[1]]]}
                },
                { type: 'month', count: 1, text: 'maand',
                 // dataGrouping: { forced: true, units:[['day',[1]]]}
                },
            ],
            buttonTheme: { // styles for the buttons
                fill: 'none',
                stroke: 'none',
                'stroke-width': 0, r: 8,
                style: {
                    color: '#37508f',
                    fontWeight: 'bold',
                    fontSize: '70%'
                },
                states: {
                    hover: {
                    },
                    select: {
                        fill: '#37508f',
                        style: {
                            color: 'white'
                        }
                    }
                    // disabled: { ... }
                }
            },
            inputPosition: { y: -23 },
            inputBoxBorderColor: 'rgba(0,0,0,0)',
            inputBoxWidth: 65,
            inputBoxHeight: 15,
            inputDateFormat: '%a %e %b %H:%M',
            inputEditDateFormat: '%e %b %H:%M',
            inputStyle: {
                color: '#37508f',
                fontWeight: 'bold',
                fontSize: '80%'
            },
            labelStyle: {
                color: '#37508f',
                fontWeight: 'bold',
                fontSize: '80%'
            },
            scrollbar: { enabled: false },
            inputEnabled: true
        },
        
          title: {
           text: BdPtitle,
            y: 40-30,
            align: 'center',
            style: {
                fontSize: '120%',
                textShadow: '1px 1px 3px #777777',
                color: '#314877'
            }
          },
          subtitle: {
           text: BdPsubtitle,
            y: 40-30+14,
            align: 'center',
            style: {
                fontSize: '110%',
                textShadow: '1px 1px 3px #777777',
                color: '#314877'
            },
          },
        chart: {
            type: 'spline',
            borderRadius: 3,
            borderWidth: 1,
            backgroundColor: '#ffffff',
            borderColor: 'silver',
            shadow: true,
            margin: [28,6,7,6],
            spacing: [25,12,10,5],
            
        },
        legend: {
            layout: 'vertical',
            borderRadius: '3px',
            backgroundColor: 'rgba(196,206,228,0.5)',
            align: 'left',
            y: 60,
            itemStyle: { fontSize: '75%', color: '#314877' },
            itemHiddenStyle: { color: '#4a6396' },
            verticalAlign: 'center',
            enabled: true,
            labelFormatter: function() {
                if( this.visible ) { return this.name ; }
                return this.name + ' (toon)';
            }
        },
        credits: { enabled: false },
        tooltip: {
            shared: true,
            valueDecimals: 1,
            dateTimeLabelFormats: {
                hour:"%a %e %b om %H uur",
                day: '%e %B %Y'
            }
        },
        xAxis: {
            labels: {
                style:{fontSize:'100%'},
                reserveSpace: false,
                step: 1
            },
            tickInterval: 7 * 86400 * 1000, // per 6 hours or per 7 days
            type: 'datetime',
            tickPosition: "outside",
            tickLength: 5,
            minorTickInterval: 2,
            minorTickLength: 22,
            minorTickPosition: 'outside',
            // minorGridLineDashStyle: 'solid',
            lineWidth: 1,
            opposite: true,
            dateTimeLabelFormats: { day:"%a %e %b", hour:"%Hh" },
            crosshair: { dashStyle: 'dot' }
        },
        yAxis: [
          {     // particals count
           title: {
                text: '\u2248 pm/uur',
                offset: 0,
                align: 'high',
                rotation: 0,
                style: {
                    fontSize: '10px',
                },
                textAlign: 'left',
                x: 5,
                y: 10,
            },
            labels: {
                format: "{value}",
                align: 'left',
                style: {
                    fontSize: '10px'
                },
                x: 5
            },
            plotLines: [{ // zero plane
                value: 0,
                color: '#BBBBBB',
                width: 1,
                zIndex: 2
            }],
            showLastLabel: false,
            maxPadding: 0.3,
            opposite: false,
            tickInterval: 1,
            gridLineColor: (Highcharts.theme && Highcharts.theme.background2) || '#F0F0F0'
          },
          { 
           title: {
                text: "\u00B5g/m\u00B3",
                offset: 0,
                align: 'high',
                rotation: 0,
                style: {
                    fontSize: '10px',
                },
                textAlign: 'right',
                x: -5,
                y: 10,
            },
            labels: {
                format: "{value}",
                align: 'right',
                style: {
                    fontSize: '10px'
                },
                x: 0,
            },
            plotBands: [
//START plotBands
                {color:"rgba(0, 32, 197, 0.05)",from: 0,to:19.5 },
                {color:"rgba(244, 230, 69, 0.15)",from: 20 ,to: 25 },
                {color:"rgba(254, 118, 38, 0.15)",from: 25.5,to: 40},
                {color:"rgba(220, 6, 16, 0.15)",from: 40.5,to:50 },
                {color:"rgba(162,23,148,0.15)",from:50.5,to:200}
//END plotBands
            ],
            showLastLabel: false,
            opposite: true,
            maxPadding: 0.3,
            tickInterval: 1,
            gridLineColor: (Highcharts.theme && Highcharts.theme.background2) || '#F0F0F0'
          }
        ],
        plotOptions: {
            series: { groupPadding: 0, borderWidth: 0.3, pointPadding: 0.03 },
            column: { shadow: true, colorByPoint: true, showInLegend: false },
            spline: { shadow: false }
        },
        series: [ 
//START series
            { type: 'spline',
              pointStart: BdPstart1,
              pointInterval: BdPunit1,
              name: BdPtitle1,
              data: BdPdata1,
              lineWidth: 1+1,
              visible: BdPvisible1,
              zIndex: 2,
              marker:{
                radius: 0
              }
            },
//END series
             ]
    });

//END DOM
});
</script>
<p>
<a id="metingen" name="metingen"></a>
In de legendum kan door aanklikken een grafiek voor een enkele fijnstof meting aan- of uitgezet worden.
Klik "toon" aan voor het aan- of uitzetten van een grafiekweergave.
<br />De chart van grafieken heeft de mogelijkheid om een bepaalde periode van de grafiek te laten zien. Met de "slider" kan in een periode in- of uitgezoomd worden. Ook kan over de tijd-as geschoven worden.
</p><br />
<p>
<!-- Highchart graphics -->
<table style='color:black;background:#f9f9f9;border-radius:7px;-moz-border-radius:7px;-webkit-border-radius:7px;box-shadow: 0px 7px 10px 0px #a0a0a0;text-align:left;padding:0px;padding-top:3px;padding-bottom:0.3%;margin:0px;border-spacing:0px;border:0px solid black;width:100%;'>
<caption valign=top align=center><span style='color:#0376ab'><a href="http://meetnetdata.rivm.nl/vuurwerk/" title="meer meetplaatsen met de fijnstof sensor van het project SamenMetenAanLuchtkwaliteit.nl het project Vuurwerk van het RIVM">Luchtkwaliteit meetresultaten</a>: de grafieken (fijnstof) metingen<br >in de regio
<!-- START regio -->
Horst a/d Maas
<!-- END regio -->
</span>.<br /><span valign=bottom style='font-size:85%;color:#2e3b4e;text-align:right'>Gebaseerd op data van o.a. <a href="https://www.luchtmeetnet.nl/stations/limburg/alle-gemeentes/alle-stoffen" title="Bezoek website met actuele metingen">het RIVM</a>
van
<!-- START first -->
zo 11 dec 01:13
<!-- END first -->
uur tot
<!-- START last -->
ma 12 dec 04:26
<!-- END last -->
uur.
</span>
</caption>
<tbody>
<tr>
<td style='padding:1%'><div title="De metingen zijn gevalideerd.Over een periode van enkele weken zijn de fijnstofmetingen aan een Chi kwadraad en Grubbs Z-score test onderworpen om de z.g. spikes, de null metingen te verwijderen en statische waarden over een periode er uit te filteren.">
<div id="VUURWERK" style="width:510px; height:300px;margin:0 auto"></div>
</div></td>
</tr>
<tr><td colspan=5 style='vertical-align:top;text-align:right;padding-right:10px;padding-bottom:6px;font-size:70%'>geactualiseerd op 
<!-- START updated -->
za 17 dec 12:41
<!-- END updated -->
</td></tr>
</tbody>
</table>
</p>
<p>
<table width=100% border=0 bordercolor=white>
<tr><td title="Gebruik de cursor om een selectie te maken van de periode en/of meer detail informatie over een meting.">
De bovenstaande tabel met de recente luchtkwaliteits metingen voor de regio
//START regio
Horst a/d Maas
//END regio
. In de grafieken worden de fijnstof metingen weergegeven van de meetstation(s) &#150;
<!-- START stations -->
Hoogheide, Horst an de Maas (PLIM)
<!-- END locations -->
&#150;
in vergelijking met de grafiek van metingen met een eenvoudige betaalbare sensor(s):
<!-- START locations -->
de Bisweide, Grubbenvorst (RIVM)
<!-- END locations -->
.
Het Vuurwerk project is momenteel afgelopen. Hoewel nog op veel plaatsen de sensor nog prima functioneert zoals in Grubbenvorst. RIVM heeft een rapportage van dit initiatief en het streven naar "Samen Meten aan Luchtkwaliteit" uitgebracht: <a href="http://www.samenmetenaanluchtkwaliteit.nl/sites/default/files/2017-03/Tijdschrift%20Lucht%20maart%202017%20-%20Samen%20meten%20aan%20luchtkwaliteit.pdf">Lucht in Onderzoek</a> in het tijdschrift Lucht (PDF 258 KB).<br />
De eenvoudige fijnstof sensor telt het aantal fijnstofdeeltjes (hoofdzakelijk PM<span style="font-size:80%">2.5</span>, een gemiddeld aantal deeltjes over een uur). De fijnstof metingen van meetstations zijn gewichtsmetingen (&micro;g/m&sup3;) per uur. 
<br />Tav de meetgegevens van goedkopere sensoren moet worden opgemerkt dat in deze meetwaarden testperiode de weergegeven waarden nog niet gekalibreerd zijn. Het zijn getelde fijnstofdeeltjes in een onduidelijke luchtstroom en geen gewichtswaarden of tellingen per luchtdeel. De sensor meetwaarden in de grafiek zijn geaggregeerd van een gemiddelde telling per 5 minuten tot een gemiddelde per uur.
<br />Bij de grafieken van de meetstations worden tevens met een achtergrondskleur de verschillende
<!-- START AQI -->
EU en WHO norm grensen (resp. 20, 25, 45 en 50 &micro;g/m&sup3;)
<!-- END AQI -->voor fijnstof weergegeven.
</td></tr>
</table>
</p>
</body>
</html>
