#!/usr/bin/perl -w

# Copyright (C) 2016, Teus Hagen, the Netherlands
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# $Id: vuurwerk.pl,v 3.13 2017/01/02 19:13:03 teus Exp teus $
my $Version = '$Revision: 3.13 $, $Date: 2017/01/02 19:13:03 $.';
$Version =~ s/\$//g;
# Description:
# script will download fijnstof sensor data from:
#       -  RIVM vuurwerk 2016 project page VW_series
# RIVM has stopped the VW project since April 2017
# data will be stored in a database table in table with name serial number
# sensor static data will be added/updated in Sensors DB table
# sensor value tables and Sensors tables will be created/updated on the fly
# in the databese DB (luchtmetingen, automatically created in needed)

# configuration constants, optional constants my be emptied.
use constant {
                                        # Default DB credentials
        USER => 'someuser', PASSWD => 'somepass', HOST => 'localhost',
        DB        => 'luchtmetingen',   # default database

        # default sensor info
        LOCATION  => 'Grubbenvorst',    # default label (optional)
        ID        => '13033927',        # default table of hourly measurements
                                        # latitude,longitude,altitude
        COORDINATES => '51.420698,6.135656', # default location coordinates (optional)
        RESOLUTION => 0.111,            # max new coordinate resolution in km
        PROJECT   => 'VW',              # default project id
        TTL       => 24*60*60,          # deactive sensor if not updated in secs

        VW_RECHECK => 12*60*60,         # sec to reload coodinates VW project
        VW_RESOLUTION => 1.0,           # max change to load new coordinates VW project
        VW_URL    => 'http://meetnetdata.rivm.nl/vuurwerk/vuurwerk-details-avg-dot.php?kit_id=VW_XXXXXXXX',
        SENSORS   => 'Sensors',         # name of sensors info table

        FALSE     => 0,
        TRUE      => 1,
        WDIR      => '/webdata/luchtmetingen/',   # working directory
};
# template to access RIVM vuurwerk (project VW) website data
# up to 30 dec 2016: http://meetnetdata.rivm.nl/vuurwerk/vuurwerk-details-avg.php?kit_id=VW_XXXXXXXX

# dependensies on Perl modules
use JSON::PP;
use Time::Piece;
use POSIX qw(strftime);
use feature "state";
use Env qw(HOME DBUSER DBPASS DBHOST);  # for working from HOME dir, and DB credits
use feature "state";            # some routines keep own state variables
use POSIX qw(strftime);
use Time::Piece;                # need to parse time strings
use autodie;
use 5.010;
use LWP::UserAgent;             # for direct http access
use URI::Escape;
use Getopt::Mixed;              # need to parse command line arguments
use LWP::UserAgent;
use DBI;

my $debug = 0;   # on debug output full HTML page to output
my $verbose = 1; # verbosity level

my $myhost   = HOST;    # mysql server
$myhost      = $DBHOST if defined $DBHOST;
my $mydb     = DB;      # mysql database name
$mydb        = $DB if defined $DB;
my $myuser   = USER;    # mysql user
$myuser      = $DBUSER if defined $DBUSER;
my $mypass   = PASSWD;  # mysql password
if( $mypass =~ /somepass/ ){
    $mypass = '';
    $mypass = $DBPASS if defined $DBPASS;
}
my $mysql;              # connection handle with mysql service

END {                   # on exit close connections
    if( $mysql and $mysql != 1 ){ $mysql->disconnect(); $mysql = 0; }
}

my  $WARNING = "WARNING:";
my  $ERROR = "ERROR:";
my  $ATTENT = "ATTENTION:";
sub isatty( ){
    no autodie; state $isatty = open(my $tty, '+<', '/dev/tty');
    return $isatty;
}

# try to hide non browser behaviour
my %agents = (
    'dotbot' => "Mozilla/5.0 (compatible; DotBot/1.1; http://www.opensiteexplorer.org/dotbot, help\@moz.com)",
    'butterfly' => "Mozilla/5.0 (Butterfly)",
    'baidu' => "Mozilla/5.0 (compatible; Baiduspider/2.0; +http://www.baidu.com/search/spider.html)",
    'slurp' => "Mozilla/5.0 (compatible; Yahoo! Slurp; http://help.yahoo.com/help/us/ysearch/slurp)",
    'google' => "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)",
    'opera' => "Opera/8.01 (Windows NT 5.1; U; nl)",
    'iphone' => "Mozilla/5.0 (iPhone; CPU iPhone OS 7_1_2 like Mac OS X) AppleWebKit/537.51.2 (KHTML, like Gecko) Version/7.0 Mobile/11D257 Safari/9537.53",
    'ipad' => "Mozilla/5.0 (iPad; CPU OS 7_1_2 like Mac OS X) AppleWebKit/537.51.2 (KHTML, like Gecko) Version/7.0 Mobile/11D257 Safari/9537.53",
    'baidu' => "Mozilla/5.0 (compatible; Baiduspider/2.0; +http://www.baidu.com/search/spider.html)",
#    'spinn' => "Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.9.0.19; aggregator:Spinn3r (Spinn3r 3.1); http://spinn3r.com/robot) Gecko/2010040121 Firefox/3.0.19",
    'ahrefsbot' => "Mozilla/5.0 (compatible; AhrefsBot/5.0; +http://ahrefs.com/robot/)",
    'msnbot' => "msnbot-media/1.1 (+http://search.msn.com/msnbot.htm)",
    'majestic' => "Mozilla/5.0 (compatible; MJ12bot/v1.4.4; http://www.majestic12.co.uk/bot.php?+)",
    'windows' => "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:18.0) Gecko/20100101 Firefox/18.0",
);

# get content from a website page: argument URL
sub get_file {
    my $file = shift;
    my $URL;
    if( not $URL ){
        $agnt = ((keys %agents)[int rand keys %agents]);
        $agnt = $agents{$agnt};
        $URL=LWP::UserAgent->new(agent => $agnt);
        $URL->timeout('40');
        # use proxy to anonimize downloads and add advertise income to these sites
        if( system("pidof privoxy >/dev/null") ){
            $URL->env_proxy;
        } else {
            $URL->proxy('http', 'http://127.0.0.1:8118/');
        }
    }
    my $resp = $URL->get($file);

    if( $resp->is_success ){
        return $resp->content;
    } else {
        print STDERR "$ERROR: Sensor data acquisition. Response status line: \n\t";
        print STDERR $resp->status_line,"\n";
    }
    print STDERR "$ERROR: No access to: $file\n" if $verbose;
    return undef;
}

# make sure we work in a well known dir
my $WDir = WDIR;
if( -d $WDir ){
   chdir( $WDir) ;
} else {
   print STDERR "Cannot chdir to working dir $WDir\n";
   $WDir = './';
}

# defaults to be redefined from command level
my $label = LOCATION;
my $identifier = ID;
my $coordinates = COORDINATES;  # only optionaly needed latitude,longitude/altitude
my $project = PROJECT;

# command options
Getopt::Mixed::init(
        'H help>H '.
        'd:i debug>d '.
        'v verbose>v '.
        'q quiet>q '.
        'p=s pass>p '.
        'u=s user>u '.
        'h=s host>h '.
        'D=s database>D '.
        'l=s label>l '.
        'i=s ID>i '.
        'n nolast>n '.
        'N nocolor>N '.
        'P=s project>P '.
        ''
);
while( my($option, $value, $arg) = Getopt::Mixed::nextOption() ){
  OPTION: {
    $option eq 'd' and do { $verbose++; $debug++;  last OPTION; };
    $option eq 'v' and do { $verbose++;  last OPTION; };
    $option eq 'q' and do { $verbose = 0;  last OPTION; };
    $option eq 'h' and do { $myhost = $value; $myhost =~ s/\s//g; last OPTION; };
    $option eq 'D' and do { $mydb = $value; $mydb =~ s/\s//g; last OPTION; };
    $option eq 'u' and do { $myuser = $value; $myuser =~ s/\s//g; last OPTION; };
    $option eq 'p' and do { $mypass = $value; last OPTION; };
    $option eq 'P' and do { $project = $value; last OPTION; };
    $option eq 'l' and do { $label = $value; $label =~ s/\s//g; last OPTION; };
    $option eq 'i' and do { $identifier = $value; $identifier =~ s/\s//g; last OPTION; };
    $option eq 'W' and do { $WDir = $value; chdir(WDir); last OPTION; };
    $option eq 'n' and do { $no_last = 1; last OPTION; };
    $option eq 'N' and do { $no_color=1; last OPTION; };
    $option eq 'H' and do {
        print STDERR <<EOF ;
        $0 [options] [arg...] (default/no argument(s):
            catch up to latest sensor sensor SN/label data of the project $project.
            Default: label/identifier $identifier.
        $0 argument uses the name, optionaly search meetnetdata.rivm.nl site for
        the name (label) as identifier.
        The info is updated into the 'Sensors' database. The downloaded data measurements
        is inserted into the database '$mydb' at table identified by project/serial nr..
        Default label or table: ${project}_${identifier} (ProjectIDSensorSerial).
        For new sensors and if the coordinate is defined the location (street,
        village) will be added in the sensors info table via Google Maps.

        Options:
 -l|--label  The measurement label to use. Default label $label
 -i|--ID        The measurement sensor serialNR to use. Default $identifier
                The database table is defined by {project_ID}_serialNR.
                Default: ${project}_$identifier

 -d|--debug     Debugging on.
 -H|--help      This help or usage message.
 -v|--verbose   Increase verbosity, default: off.
 -q|--quiet     Be as quiet as possible, turns verbose off, default: off.

 -D|--database  The Database name to use (default: $mydb).
 -h|--host      The DB host for DB access (default: $myhost).
 -u|--user      The DB user for DB access (default: $myuser).
 -p|--pass      The DB user password for DB access (default: X?x?X).
 -P|--project   The project name (default: VW). Is used as component in data DBtable name.
 -n|--nolast    Update DB with all values found and not from last value found.
                Default: update from last value found.
 -N|--nocolor   Use no coloring in error/warning/attention messages.
 -W|--wdir      The working directory with eg rrd database in rrd_data directory.

$Version
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License V4 as published by
the Free Software Foundation.
EOF
        exit(0);
        };
   };
}
Getopt::Mixed::cleanup();
$verbose = 3 if $debug;

die "FATAL $ERROR: no database user/password defined, e.g. define use environment variable DBUSER and/or DBPASS\n" if length($mypass) <= 0;


# establish DB connection
sub Check_DB { # called only once from first call to query routine
    state $checkedOnce = 0;
    if( $mysql ){ return 1; }
    return 0 if length( $mydb ) < 1;
    return 0 if length( $myhost ) < 1;
    return 0 if length( $myuser ) < 1;
    return 0 if length( $mypass ) < 1;
    $mysql = DBI->connect("DBI:mysql:".$mydb.":".$myhost,$myuser,$mypass);
    if( not $mysql ){
        if( $checkedOnce > 0 ){
            print STDERR "$ERROR Cannot open mysql database ".$mydb." on host ".$myhost.", with user ".$myuser.", error: ". DBI->errstr."\n" ;
            $mysql = 1; return 0;
        }
        print STDERR "$ERROR Cannot open mysql database ".$mydb." on host ".$myhost.", with user ".$myuser.", error: ". DBI->errstr."\nWill try to create it\n" ;
        $checkedOnce = 1;
        Create_DB();
        return Check_DB();
    }
    return 1;
}

# DB columns for various types of sensors
%DB_table = (
             'id' => "id TIMESTAMP DEFAULT CURRENT_TIMESTAMP COMMENT 'datum/tijd laatste verandering sensor waarde'",
            'datum' => "datum datetime UNIQUE default NULL COMMENT 'datum/tijd UNIX time stamp measurement'",
            'pm' => "pm DECIMAL(5,2) default NULL COMMENT '5 minutes avg particular count seen by sensor'",
                        'so' => "so DECIMAL(5,2) default NULL COMMENT 'zwavel oxyde ug/m3'",
            'so2' => "so2 DECIMAL(5,2) default NULL COMMENT 'zwavel dioxyde ug/m3'",
            'no' => "no DECIMAL(5,2) default NULL COMMENT 'stikstof oxyde ug/m3'",
            'no2' => "no2 DECIMAL(5,2) default NULL COMMENT 'stikstof dioxyde ug/m3'",
            'nh3' => "nh3 DECIMAL(6,2) default NULL COMMENT 'ammonium ug/m3'",
            'o3' => "o3 DECIMAL(5,2) default NULL COMMENT 'ozon, tri-oxide ug/m3'",
            'pm_10' => "pm_10 DECIMAL(5,2) default NULL COMMENT 'fijnstof 10 ug/m3'",
            'pm_25' => "pm_25 DECIMAL(5,2) default NULL COMMENT 'fijnstof 2.5 ug/m3'",
            'roet' => "roet DECIMAL(5,2) default NULL COMMENT 'roet (soot) ug/m3'",
            'benzeen' => "benzeen DECIMAL(5,2) default NULL COMMENT 'benzeen ug/m3'",
            'tolueen' => "tolueen DECIMAL(5,2) default NULL COMMENT 'tolueen ug/m3'",
            'wr' => "wr SMALLINT(6) default NULL COMMENT 'windrichting 0-360 graden'",
            'ws' => "ws DECIMAL(5,2) default NULL COMMENT 'windsnelheid m/sec'",
            'temp' => "temp DECIMAL(5,2) default NULL COMMENT 'temperatuur oC'",
            'rv' => "rv DECIMAL(5,2) default NULL COMMENT 'relatieve vochtigheid'",
            'luchtdruk' => "luchtdruk INT(11) default NULL COMMENT 'luchtdruk mbar'",
            # sensor value is valid (will be switched of by statistical filters
            'name_valid' => "pm_valid BOOL default 1 COMMENT 'PM value is valid'",
);

# table name composition: {project}_{serial}
my $DB_table_labels = "
            coordinates varchar(25) default NULL COMMENT 'geolocation: altitude,longitude,altitude',
            label varchar(50) default NULL COMMENT 'label name',
            street varchar(50) default NULL COMMENT 'location road name',
            village varchar(50) default NULL COMMENT 'location village name',
            municipality varchar(50) default NULL COMMENT 'lowest level government',
            province varchar(50) default NULL COMMENT 'second level government',
            sensors varchar(50) default NULL COMMENT 'sensor configuration a list of IDs',
            description varchar(100) default NULL COMMENT 'description/comments',
            first datetime default '2001-01-01 00:00:00' COMMENT 'first activation',
            active BOOL default 1 COMMENT 'sensor is active',
            last_check datetime default now() COMMENT 'last date/time check location',
            project varchar(10) default 'VW' COMMENT 'project id, table id',
            serial varchar(15) default NULL COMMENT 'identifier'
";

my $DB_sensor_types = "
            producer varchar(25) default NULL COMMENT 'producer of the sensor'.
            type varchar(50) default NULL COMMENT 'type number',
            pollutant varcahr(10) default NULL COMMENT 'DB pollutant identifier'
";
        

# create new database and add standard tables: main table, dayly averages, days
sub Create_DB {
    my $string = "
        CREATE DATABASE  IF NOT EXISTS $mydb;
        USE $mydb;\n";
    # create a table to be able to finds dates and ranges in main table

    open(my $MYSQL, "| /usr/bin/mysql -u $myuser -p$mypass -h $myhost" ) || die "FATAL $ERROR mysql create DB $mydb: $!\n";
    print $MYSQL $string;
    close $MYSQL;
    foreach ( $project .'_'.$identifier, 'Sensors', 'Types' ){ Create_TBL( $_ ); }
    return 1;
}

# obtain all sensors for a station in the DB table
my %DB_cols = ( );      # sensors names per table
sub Get_Sensors {
    my $tbl = shift; return 0 if not $tbl;
    Check_Tbl($tbl) if not $DB_cols{$tbl};
    if( not defined $DB_cols{$tbl}{"id"} ){
        # only once at the start we build a column existance cache
        my $qr = query($tbl, "DESCRIBE $tbl;");
        if( (not $qr) || ($#{$qr} < 0) ){
            print STDERR "$ERROR Cannot obtain table $tbl description\n";
            return 0;
        }
        # mysql counts cells from 0!
        $DB_nr_cols{$tbl} = 0 if not defined $DB_nr_cols{$tbl};
        for( my $index = 0; $index <= $#{$qr} ; $index++ ){
            $DB_cols{$tbl}{ $qr->[$index] } = 1;
            next if $qr->[$index] =~ /^(id|datum|.*_(valid|ppb))$/ ;
            $DB_nr_cols{$tbl}++;
            $DB_cols{$tbl}{$qr->[$index]} = 1;
        }
    }
    return 1;
}

# add some comment to a DB COMMENT string
sub Add_comment {
    my $cmt = shift; $cmt = '' if not defined $cmt;
    my $add = shift; $add = '' if not defined $add;
    if( $cmt =~ /.*COMMENT '/ ) {
        $cmt =~ s/COMMENT '([^']*)/COMMENT '$1$add/;
    }
    return $cmt;
}

# check if sensor is available in table if not add the column
sub Check_Sensor {
    my $tbl = shift;
    my $col = lc(shift);
    my $comment = shift;
    return 0 if not Get_Sensors($tbl);
    $comment = "unkown name $col" if ( (not defined $comment) || (not $comment) );
    $comment .= strftime(", added: %Y/%m/%d", localtime(time));

    return 0 if defined $DB_cols{$tbl}{$col};
    $rate .= ", "."ADD COLUMN ".$col.$DB_table{'name_valid'};
    $rate .= ", "."ADD COLUMN ".$col.$DB_table{'name_ppb'} if $col =~ /^(so2|no|no2|nh3|o3|benzeen|c6h[2-9][ch2-9]*)$/;;
    if( $DB_table{$col} ){
        print STDERR "$ATTENT Adding column $col to table $tbl\n" if $verbose > 0;
        my $cmt = Add_comment($DB_table{$col},
            strftime(", added: %Y/%m/%d", localtime(time)) );
        query($tbl, "ALTER TABLE $tbl ADD COLUMN $cmt, $rate;");
    } else {
        print STDERR "$ATTENT Adding unknown new column $col to table $tbl\n";
        query($tbl, "ALTER TABLE $tbl ADD COLUMN $col DECIMAL(5,2) default NULL COMMENT '$comment', $rate;");
    }
    $DB_cols{$tbl}{$col} = -1;
    $DB_nr_cols{$tbl}++ if $col !~ /^(id|datum|.*_(valid|ppb))$/ ;
    return 0;
}

# one row of values from one column out of the DB table
sub query {
    return 0 if Check_DB == 0;
    my $tbl = shift; Check_Tbl($tbl) if $tbl;
    my $q = shift;
    if( (not $mysql) or ($mysql == 1) or (not $q) ){ return undef; }
    print STDERR "MYSQL: $q\n" if $debug > 1;
    my $sth = $mysql->prepare($q); $sth->execute();
    #  ref to 2 dimensional refs to array string values
    if( $q =~ /^\s*(show|describe|select)/i ){
        my $r = $sth->fetchall_arrayref();
        print STDERR "$ERROR mysql query: $q with error:\n" . DBI->errstr . "\n"
            if ( DBI->errstr &&  $#{$r} < 0 );
        return undef unless $r->[0];
        my @rts = ();
        for( my $i=0; $i <= $#{$r}; $i++ ){
            $rts[$i] = $r->[$i][0] if $#{$r->[$i]} >= 0;
        }
        if( not @rts ){
            # print STDERR "$WARNING mysql query: $q with empty array:\n" . DBI->errstr . "\n";
        }
        $sth->finish(); return \@rts;
    } else { return (DBI->errstr ? undef : 1); }
}
# matrix of values
sub long_query {
    return 0 if Check_DB == 0;
    my $tbl = shift; Check_Tbl($tbl) if $tbl;
    my $q = shift;
    if( (not $mysql) or ($mysql == 1) or (not $q) ){ return undef; }
    print STDERR "MYSQL: $q\n" if $debug > 1;
    my $sth = $mysql->prepare($q); $sth->execute();
    #  ref to 2 dimensional refs to array string values
    if( $q =~ /^\s*(select)/i ){
        my $r = $sth->fetchall_arrayref();
        return undef unless $r->[0];
        if( $#{$r} < 0 ){
            print STDERR "$ERROR mysql query: $q with error:\n" . DBI->errstr . "\n";
        }
        $sth->finish(); return $r;
    } else { return undef; }
}

# collect from database the table column names
# if table does not exists create it
sub Check_Tbl {
    my $tbl = shift; my $noprobe = shift; # probe if it exists only
    return 0 if not $tbl;
    return 1 if defined $DB_cols{$tbl};
    my $qry = query('',"SHOW TABLES");
    for( my $i = 0; $i <= $#{$qry}; $i++ ){
        # next if $qry->[$i] !~ /^$mytbl/;
        $DB_cols{$qry->[$i]} = {} if not $DB_cols{$qry->[$i]}; # the table already exists
    }
    return 1 if defined $DB_cols{$tbl};
    return 0 if (defined $noprobe) && not $noprobe;
    Create_TBL( $tbl );
    $DB_cols{$tbl} = {};
    return 1;
}

# create new table if not existing
sub Create_TBL {
    my $tbl = shift;
    my $comment = '';
    if( $label ){
        my $mydate = strftime("%Y/%m/%d", localtime(time));
        $comment = "COMMENT='Sensor with Label $label, init: $mydate'";
    }
    return 0 if not $tbl;
    my $string = "CREATE TABLE $tbl (";
    foreach ('id', 'datum' ){
        $string .= "\t$DB_table{$_},";
    }
    if ( $tbl !~ /\_/ ) {
        $string .= $DB_table_labels if $tbl =~ /Sensors/;
        $string .= $DB_sensor_types if $tbl =~ /Types/;
    }
    $string =~ s/,\s*$//;
    $string .= ") ENGINE=InnoDB DEFAULT CHARSET=latin1 $comment;\n";
    $DB_cols{$tbl} = {};
    query( '', $string );
    return 1;
}

# remove once the spike.
# We use 3 X stddev within a window last 2 days as Chi square test
# maybe use clustering techniques or learning technique as Markov chaining
sub remove_peaks {
    my ($TBL,$max) = @_;
    my $dev = $max; $dev =~ s/.*\+//; $max =~ s/\+.*//;
    my $min = $max - $dev; $max += $dev; $min = 0 if $min < 0;
    query($TBL,"UPDATE $TBL SET pm_valid = 0 WHERE
                (pm >= $max) OR (pm <= $min)");
}

# get the avg and sigma value to detect peaks
# moving window is 2 days
sub get_bandwidth {
    my ($TBL,$sensor,$diff,$last) = @_; $last = 0 if not defined $last;
    state $S = ''; state $L = 0; state $A;
    $sensor = 'pm' if not defined $sensor;
    $diff = 2 if (not defined $diff) || (not $diff); # default 2 days window
    @MinMax = (-10000,10000);
    $A = \@MinMax if not defined $A;
    return $A if ($S eq $sensor) && (abs($L - $last) < 60*60); # cached for 1 hour
    # check if the are enough values to start peak detection
    my $rslt = query($TBL,"SELECT count(*) FROM $TBL WHERE ${sensor}_valid");
    return \@MinMax if (not defined $rslt) || ($#{$rslt} < 0);
    return \@MinMax if $rslt->[0] < 12*24*$diff; # minimal 2 days of measurements
    my $cnt = $rslt->[0];
    if( not $last ) {
        $rslt = query($TBL,"SELECT UNIX_TIMESTAMP(datum) FROM $TBL WHERE not isnull(${sensor}) ORDER BY datum DESC LIMIT 1");
        return \@MinMax if $#{$rslt} < 0;
        $last = $rslt->[0]
    }
    $rslt = query($TBL,"SELECT CONCAT(avg(${sensor})-2.5*stddev(${sensor}),'=',avg(${sensor})+2.5*stddev(${sensor}))
                FROM $TBL WHERE not isnull(${sensor}) AND
                    (datediff(datum,FROM_UNIXTIME($last)) <= $diff) AND ${sensor}_valid");
    return \@MinMax if (not defined $rslt) || ($#{$rslt} < 0);
    $MinMax[0] = $rslt->[0]; $MinMax[0] =~ s/=.*//;
    $MinMax[0] = 0 if $MinMax[0] < 0;
    $MinMax[1] = $rslt->[0]; $MinMax[1] =~ s/.*=//;
    $S = $sensor; $L = $last; $A = \@MinMax;
    printf STDERR ("Bandwidth on table $TBL %d days up to %s: %4.1f - %4.1f\n",
            $diff,strftime("%Y-%m-%d %H:%M",localtime($last)), $MinMax[0], $MinMax[1])
        if $verbose > 1;
    return $A;
}

# a Chi square test on the value of a sensor in a window of 2 days of values
# returns the validation
sub Chi_square {
    my ($TBL,$sensor,$value, $date) = @_;
    $date = time if not defined $date;
    my $MinMax = get_bandwidth( $TBL,$sensor, 2, $date);
    return 1 if ($value < $MinMax->[1]) && ($value > $MinMax->[0]);
    return 0;
}
    
# add sensor data to database; Args: table_name, UNIX_timestamp, ref array
# array has [pollutant_name, pollutant_value, [value is valid]]
sub Add_sensed{
    my ($TBL, $date, $sensors) = @_;
    my $strg = '';
    state $err_msg = 0;
    # state $once = 0;
    return 0 if not $date;
    return 0 if $#{$sensors} < 0;
    
    for( my $i = 0; $i <= $#{$sensors}; $i++ ) {
        next if ($#{$sensors->[$i]} < 1); # should contain a triple value
        my $sensor = $sensors->[$i][0]; next if not $sensor;
        Check_Sensor($TBL,$sensor);
        my $valid; my $value = $sensors->[$i][1]; 
        if( defined $sensors->[$i][2] ) {
            $valid = $sensors->[$i][2];
        } else {
            $valid = Chi_square($TBL,$sensor,$value,$date); # validate the value
        }
        if( $value !~ /^[0-9\-\.]+$/ ){
            printf STDERR ("$WARNING: at %s sensor for $sensor sensed $value is not a value!\n",
                strftime("%Y/%m/%d %H:%M",localtime($date)) )
              if not $err_msg++;
            $value = 'NULL';
        }
        $valid = 0 if (not $value) && ($sensor !~ /(temp|wr|ws)/);
        $strg .= ", ${sensor} = ${value}, ${sensor}_valid = $valid";
    }
    $strg =~ s/^,//; 
    return 0 if not $strg;
        
    # if needed set this in your mysql init file
    # next will fail for normal DB users
    #if ( not $once ) {
    #    query('SET GLOBAL @@sql_mode=(SELECT REPLACE(@@sql_mode,"ONLY_FULL_GROUP_BY",""))');
    #    $once++;
    #}
    # first a placeholder
    query($TBL,"REPLACE INTO $TBL (datum) VALUES (FROM_UNIXTIME($date));");
    $strg = "UPDATE $TBL SET $strg, id = now()  WHERE datum = FROM_UNIXTIME($date)";
    query($TBL,$strg);
    return 1;
}

# get last date of entered values of the sensor
sub Last_date {
    return 0 if (defined $no_last) && (not $no_last);
    my $table = shift;
    my $rslt = query($table, "SELECT UNIX_TIMESTAMP(datum) FROM $table ORDER BY datum DESC LIMIT 1");
    return 0 if (not defined $rslt) || ($#{$rslt} < 0);
    return $rslt->[0];
}

# calculate distance between two locations,
# switch lat/long only OK for northern world part
sub distance {
    my ($lat1, $lon1, $lat2, $lon2, $unit) = @_;
    for( my $i = 0; $i < $#_; $i++ ) { return 0 if not defined $_[$i]; }
    if ( $lat1 < $lon1 ) { # lat/long sequence
        my $t = $lat1; $lat1 = $lon1; $lon1 = $t;
    }
    if ( $lat2 < $lon2 ) { # lat/long sequence
        my $t = $lat2; $lat2 = $lon2; $lon2 = $t;
    }
    $unit = 'K' if not defined $unit;
    my $theta = $lon1 - $lon2;
    my $dist = sin(deg2rad($lat1)) * sin(deg2rad($lat2))
        + cos(deg2rad($lat1)) * cos(deg2rad($lat2)) * cos(deg2rad($theta));
    $dist  = acos($dist);
    $dist = rad2deg($dist);
    $dist = $dist * 60 * 1.1515;
    if ($unit eq "K") {
      $dist = $dist * 1.609344;
    } elsif ($unit eq "N") {
      $dist = $dist * 0.8684;
    }
    return ($dist);
}

# get the arccos function using arctan function
sub acos {
    my ($rad) = @_;
    my $ret = atan2(sqrt(1 - $rad**2), $rad);
    return $ret;
}

# converts decimal degrees to radians
sub deg2rad {
    my ($deg) = @_;
    my $pi = atan2(1,1) * 4;
    return ($deg * $pi / 180);
}

# converts radians to decimal degrees
sub rad2deg {
    my ($rad) = @_;
    my $pi = atan2(1,1) * 4;
    return ($rad * 180 / $pi);
}

# print distance(51.420658, 6.135463, 51.4192233, 6.147437) . " Km\n";
# should give 0.845447426014348 Km

# get location info for a sensor and update info if location is changed
sub VW_Info {
    my ($serial) = @_;
    my %info; my $qry; my $indx = '';
    return undef if (not defined $serial) || (not $serial);
    my $sensors = SENSORS;
    return undef if not Get_Sensors($sensors);
    my @DBqry = keys %{$DB_cols{$sensors}};
    push @DBqry, 'datum'; # datum last in array
    for( my $i = 0; $i <= $#DBqry; $i++ ) {
        $DB_cols{$sensors}{$DBqry[$i]} = $i; # columns are now sorted for this use
        $DBqry[$i] = "UNIX_TIMESTAMP($DBqry[$i])" if $DBqry[$i] =~ /(id|datum|first|last_check)/;
    }
    $qry = long_query($sensors,'SELECT '. join(',',@DBqry) ." FROM $sensors
            WHERE  serial = '$serial' OR label = '$serial'
            ORDER BY datum DESC, id DESC");
    my $new = ( ($#{$qry} < 0) || (not $qry->[0]) || ($#{$qry->[0]} < 0) ) ? 1:0;
    my $renew = 0;
    @DBqry = @{$qry->[0]};
    # check 2 times a day
    $renew = 1 if (not $new)
        && ($DBqry[$DB_cols{$sensors}{last_check}] < (time - VW_RECHECK));
    if ( $new || $renew ) {
        # unknown identifier so try to collect it from main website
        $data = VW_URL; $data =~ s/vuurwerk-details.*//;
        $data = get_file( $data );
        # can be obtained by visiting: http://meetnetdata.rivm.nl/vuurwerk/
        # addMarker(51.5675, 5.0837, '<table><tr>
        # <td class=oranje>sensor
        # <td class=blauw>VW_13018654<tr>
        # <td class=oranje>label 
        # <td class=blauw>Tilburg
        # <td class=oranje>laatste meting in data
        # <td class=blauw>2016-12-15 12:27:27</table>
        #       <a class="info dlink" href=./vuurwerk-details-avg.php?kit_id=VW_13018654>&lt;grafiek 5-minuutgemiddelden&gt;<a>
        #       <br>', '<div class=tooltipnaam>VW_13018654<br>Tilburg</div>', 'icon_rood');
        # marker type since 24-12-2016
        # addMarker(51.4207, 6.13566, '<table><tr>
        # <td class=oranje>sensor
        # <td class=blauw>VW_13033927<tr>
        # <td class=oranje>label 
        # <td class=blauw>Grubbenvorst</table>
        #       <a class="info dlink" href=./vuurwerk-details-avg-dot.php?kit_id=VW_13033927>grafiek ruwe data<a>
        #       <br><a class="info dlink" href=./vuurwerk-details-afgeleide-gem.php?kit_id=VW_13033927>grafiek afgeleide data<a><br>', '<div class=tooltipnaam>VW_13033927<br>Grubbenvorst</div>', 'icons[9]');
        my @lines = split /\n/, $data;
        for ( my $i = 0; $i <= $#lines; $i++ ) {
            # got all sensors appearing on a map
            next if $lines[$i] !~ /addMarker/;
            next if $lines[$i] !~ /$serial/i;
            while ( ($lines[$i] !~ /icon/) && ($i <= $#lines) ){
                chop $lines[$i++]; $lines[$i] = "$lines[$i-1] $lines[$i]";
            }
            my @rec = split /<td\s/, $lines[$i];
            for( my $j = 0; $j <= $#rec; $j++ ) {
                                                        # collect the info for this label
                if ( $rec[$j] =~ /addMarker.([0-9\.]+)[,\s]+([0-9\.]+).*/ ) {
                    # RIVM coordinates have max 4 decimals: resolution 111 m
                    my $long = $1; my $lat = $2;  # coordinates
                    $info{coordinates} = "$long,$lat"; # may add altitude
                } elsif ( $rec[$j] =~ /VW_([0-9]+)<tr>/ ) {
                    $tid = $1; $identifier = $tid;      # identifier
                    $info{serial} = $tid;
                } elsif ( $rec[$j] =~ />label/ ) {
                    $j++; $rec[$j] =~ s/[^>]*>([^<]*)<.*/$1/;
                    $label = $rec[$j];                  # label or label name
                    $info{label} = $label;
                } elsif( $rec[$j] =~ /laatste meting/ ) {
                    my $last = $rec[$j+1]; $last  =~ s/[^>]*>([^<]*)<.*/$1/;
                    $info{last} = $last; # will be updated with date/time json data
                } elsif( $rec[$j] =~ /'(icon.*)'/ ) {
                    # active indicator gives inactive,relative low-high rate to rest
                    # will be updated from json data as well: active = 1
                    my $active = $1;
                    if( $rec[$j] =~ /href=["']?(\.\/[^=]*)=/ ){
                        # check if RIVM changed their behavior
                        my $nurl = $1;
                        if( index(VW_URL,$nurl) < 0 ){
                            printf STDERR ("ATTENTION: url to details changed from %s to %s!\n",
                                VW_URL,$nurl);
                        }
                    }
                    if( $active =~ /icon_([a-z]+)[^a-z]*/ ) {
                        # indicator from before 24-12-2016: color mostly red
                        $active = $1; 
                        if( $active =~ /(rood)/ ) { $active = 1; }
                        else { $active = 0; }
                    } elsif( $active =~ /icon\[([0-9]+)\]/ ) {
                        $active = $1; $active = 1 if $active;
                    } else {
                        print STDERR "$ATTENT: found different active notation: $active\n";
                        $active = 0;
                    }
                    $info{active} = $active;
                    last;
                }
            }
            last;
        }
        if ( $renew && (defined $info{coordinates}) ) {
            #  hope both coordinates are on same quarter
            my @prev = split /[,;\s]/, $DBqry[$DB_cols{$sensors}{coordinates}];
            my @cur = split /[,;\s]/, $info{coordinates};
            # update info to new coordinate if distance > VW_RESOLUTION
            $renew = $new = 0 if distance($prev[0],$prev[1],$cur[0],$cur[1]) < VW_RESOLUTION;
            if( $renew ){
                foreach my $key ('street','village','municipality','province') {
                    $info{$key} = undef;
                }
            }
        }
        query($sensors,"UPDATE $sensors SET active = 0 WHERE serial = '$info{serial}'") if $renew;
        query($sensors,"INSERT INTO $sensors (serial,label,coordinates,datum,active)
            VALUES('$info{serial}','$info{label}','$info{coordinates}',". ($info{active} ? 1 : 0).")")
          if $new || $renew ;
        printf STDERR ("Found new sensor:\n\tid: %s\n\tlabel: %s\n\tgeolocation(altitude,longitude): %s\n\tlast time seen: %s\n\tactive mark: %s\n",
            (defined $info{serial} ? $info{serial} : 'none'),
            (defined $info{label} ? $info{label} : 'none'),
            (defined $info{coordinates} ? $info{coordinates} : 'none'),
            (defined $info{last} ? $info{last} : 'none'),
            (defined $info{active} ? $info{active} : 'none')) if $new;
        printf STDERR ("$ATTENT: Sensor serial %s location was: %s, moved to %s\n",
            (defined $info{serial} ? $info{serial} : 'none'),
            (defined $DBqry[$DB_cols{$sensors}{coordinates}] ? $DBqry[$DB_cols{$sensors}{coordinates}] : 'none'),
            (defined $info{coordinates} ? $info{coordinates} : 'none') )
          if $renew && $verbose;
        
    }
    if( not $new ) {    # get missing info from database info table
        foreach my $key (keys %{$DB_cols{$sensors}}) {
            next if $key =~ /(id)/;
            next if defined $info{$key};
            next if not defined $DBqry[$DB_cols{$sensors}{$key}];
            if ( $key =~ /datum/ ) {
                $info{last} = $DBqry[$DB_cols{$sensors}{$key}];
            } else {
                $info{$key} = $DBqry[$DB_cols{$sensors}{$key}];
            }
        }
        query($sensors,"UPDATE $sensors SET last_check = now()
            WHERE serial = '$info{serial}' and coordinates = '$info{coordinates}'");
    }
    if( $renew || ((defined $info{coordinates}) && (not defined $info{street})) ) {
        my %address = Get_address($info{coordinates}); my $qry = '';
        foreach my $key ('street','village','municipality','province') {
            $address{$key} = '' if not defined $address{$key};
            $info{$key} = $address{$key};
            if( not $address{$key} ) {
                $address{$key} = 'null'
            } else { $address{$key} = "'$address{$key}'"; }
            $qry .= ', ' if $qry; $qry .= "${key} = $address{$key}"; 
        }
        query($sensors,"UPDATE $sensors SET $qry
            WHERE serial = '$info{serial}' AND coordinates = '$info{coordinates}'")
          if $qry;
    }
    return %info;
}

# collect info from database SENSORS table. Return hash table with info
sub Get_Info {
    my $id = shift; return undef if (not defined $id) || (not $id);
    my $qry; my %info; my $indx = '';
    my $sensors = SENSORS; Check_Tbl($sensors);
    foreach my $i ('serial','label') {
        $qry = query($sensors,"SELECT count(*) FROM $sensors WHERE  $i = '$id'");
        if( $qry->[0] ) {
            $indx = $i; $info{$i} = $id; last;
        }
    }
    return %info if not $indx;
    $qry = query($sensors,"SELECT
            CONCAT(serial,';', label,';', coordinates,';', UNIX_TIMESTAMP(first),';',UNIX_TIMESTAMP(last_check))
            FROM $sensors WHERE $indx = '$id' ORDER BY datum DESC");
    return undef if $#{$qry} < 0;
    my @in = split /;/, $qry->[0];
    $info{serial}=$in[0]; $info{label}=$in[1]; $info{coordinates}=$in[2];
    $info{first}=$in[3];
    return %info;
}

# Googl maps reverse conversion coordinate -> human address
sub CheckType {
    my ($types,$type) = @_;
    for( my $i = 0; $i <= $#{$types}; $i++ ){
        return 1 if $types->[$i] =~ /^$type$/;
    }
    return 0;
}

# return a hash table with street(+nr), village, municipality, province
sub Get_address {
    my $LatLng = shift; my %address;
    return %address if not $LatLng;
    my $Gmaps = 'https://maps.googleapis.com/maps/api/geocode/json?latlng=';
    $LatLng =~ s/\s//g if $LatLng =~ /[;,]/;
    my @coord = split /[;,\s]/, $LatLng;
    if ( $coord[0] < $coord[1] ) {
        my $tmp = $coord[0]; $coord[0] = $coord[1]; $coord[0] = $tmp;
    }
    $LatLng = join ',', @coord;
    my $json = get_file($Gmaps . $LatLng);
    return %address if not $json;
    my $struct = decode_json($json);
    return %address if (not defined $struct->{status})
        || ($struct->{status} !~ /OK/i) || (not defined $struct->{results} );;
    my $nr; my $street = 'null'; my $municipality = 'null'; my $province = 'null';
    my $village = 'null';
    for( my $j = 0; $j <= $#{$struct->{results}}; $j++ ){
        next if not defined $struct->{results}[$j]{address_components};
      for( my $i = 0; $i <= $#{$struct->{results}[$j]{address_components}}; $i++) {
print STDERR "$struct->{results}[$j]{address_components}[$i]{short_name}\n";
        if( CheckType($struct->{results}[$j]{address_components}[$i]{types},'route') ){
           $street = "$struct->{results}[$j]{address_components}[$i]{short_name} $street" ;
           $street =~ s/\snull//;
        } elsif( CheckType($struct->{results}[$j]{address_components}[$i]{types},'street_number') ){
            $street = "$street $struct->{results}[$j]{address_components}[$i]{short_name}";
            $street =~ s/null\s//;
        } elsif( CheckType($struct->{results}[$j]{address_components}[$i]{types},'locality') ) {
            $village = $struct->{results}[$j]{address_components}[$i]{short_name};
        } elsif( CheckType($struct->{results}[$j]{address_components}[$i]{types},'administrative_area_level_2') ){
            $municipality = $struct->{results}[$j]{address_components}[$i]{short_name};
        } elsif( CheckType($struct->{results}[$j]{address_components}[$i]{types},'administrative_area_level_1') ){
            $province = $struct->{results}[$j]{address_components}[$i]{long_name};
        }
      }
      last if $street !~ /null/;;
    }
    $address{street} = $street;
    $address{village} = $village;
    $address{municipality} = $municipality;
    $address{province} = $province;
    return %address;
}

# my %address = Get_address('51.419652,6.146031');

# for sensors published by RIVM (vuurwerk 2016/17 project)
# download json data from sensor arg0: identifier if not a number look it up
# and detect the identifier from main website. Use identifier as table name
# TO DO: detect most nearby sensor for a coordinate
# data acquision routine is dedecated to this VW (RIVM 2016) project
sub VW_get_store {
    my $tid = shift;
    my $data;
    my $sensor = 'pm';          # default sensor type
    my $sensors = SENSORS;      # the sensors info table name
    $tid =~ s/VW_//;
    my %info = VW_Info($tid);
    if ( (not defined $info{serial}) || (not $info{serial}) ) {
        print "$WARNING: Cannot find info for serial/location $tid in VM project.\n";
        return undef;
    }
    my $URL = VW_URL; $URL =~ s/XXXXXXXX//;
    $data = get_file( $URL . $info{serial});
    return undef if not $data;
    my @lines = split /\n/, $data; $data = '';
    for( my $i = 0; $i <= $#lines; $i++ ) {
        next if ($lines[$i] !~ /var\s+data\s+=\s*\[/) && (not $data);
        $data .= $lines[$i]; $data .= "\n";
        last if ($lines[$i] =~ /\]\s*;/);
    }
    return undef if not $data;
    # get the json array of hash objects
    $data =~ s/^.*var\s+data\s+=\s+\[\s*/\[/m; $data =~ s/\s*\]\s*;.*/\]/m;
    {   # seems RIVM is not fully JSON conformant 
        my $json = JSON::PP->new->allow_barekey(1); $json->allow_singlequote(1);
        $data = $json->decode($data);
    }
    # array elelement record format: {x: '2016-12-12 22:05:00', y: 0.09, group: 2},
    $info{count} = 0;
    my $last = Last_date('VW_'.$info{serial});
    my $frst = 0;
    for ( my $i = 0; $i <= $#{$data}; $i++ ){
        next if (not defined $data->[$i]{x}) || (not $data->[$i]{y}) ;
        $info{last} = Time::Piece->strptime($data->[$i]{x} . " +0100", "%Y-%m-%d %H:%M:%S %z")->epoch;
        next if $info{last} <= $last;
        my @triple;
        # only one pollutant sensed
        push @triple, ['pm',$data->[$i]{y}]; # pollutant, value, optional validated
        if (Add_sensed( 'VW_'.$info{serial}, $info{last} , \@triple) ){
            $info{count}++;
            $info{first} = $info{last} if not defined $info{first};
            $frst = $info{from} = $info{last} if not $frst;
        }
    }
    my $active = 1;
    $active = 0 if (defined $info{active}) && (not $info{active});
    $active = 0 if $info{last} < (time - TTL);
    $info{active} = ($active?'yes':'no');
    # and update the station/location record
    query($sensors,"UPDATE $sensors SET first = FROM_UNIXTIME($frst)
        WHERE serial = '$info{serial}' AND coordinates = '$info{coordinates}'
            AND first = '2001-01-01 00:00:00'")
      if $frst;
    query($sensors,"UPDATE $sensors SET datum = FROM_UNIXTIME($info{last}),
        active = $active
        WHERE serial = '$info{serial}' AND coordinates = '$info{coordinates}'");
    return \%info;
}

# provide colored error output
if( (not defined($no_color)) && isatty() ){
    $WARNING = "\033[1;35mWARNING\033[0m:";
    $ERROR = "\033[1;31mERROR\033[0m:";
    $ATTENT = "\033[1;32mATTENTION\033[0m:";
}

my $INFO;
my $SENSOR = '';

# walk over the command line defined arguments to collect sensor values
# default argument is PROJECT and sensor ID: $identifier
$SENSOR = shift @ARGV if defined $ARGV[0];
$SENSOR = $identifier if not $SENSOR;
while( defined $SENSOR ) {
    if( $SENSOR =~ /^VW_/i ){
        $project = 'VW'; $SENSOR =~ s/VW_//i;
    }
    # switch to collect  and store data routine for the project sensor
    if ( $project eq 'VW' ){ # vuurwerk 2016/17 RIVM project
        # $INFO = VW_get_store($SENSOR); # stopped since 27 April 2017
	$INFO = undef
    }
    if( not defined $INFO ) {
        print STDERR "Sensors data acquisition for $SENSOR, project $project FAILED\n"
	    if $verbose;
        $SENSOR =  shift @ARGV;
        next;
    }
    printf STDERR ("Found data for sensor id %s %s%s:\n",
            $project, (defined $INFO->{serial} ? $INFO->{serial} : '?'),
            (defined $INFO->{label} ? ' ('.$INFO->{label}.')' : ''))
        if $verbose;
    printf STDERR ("\tcoordinates: %s\n\tinstalled %s\n\tcurrent active mark: %s\n",
            (defined $INFO->{coordinates} ? $INFO->{coordinates} : 'unknown'),
            (defined $INFO->{first} ? strftime("%Y-%m-%d %H:%M",localtime($INFO->{first})) : 'first'),
            (defined $INFO->{active} ? $INFO->{active} : '?'))
        if $verbose > 1;
    printf STDERR ("\tadded : from %s up to %s in total %d values\n",
            ((defined $INFO->{from}) && $INFO->{from} ? strftime("%Y-%m-%d %H:%M",localtime($INFO->{from})) : '?'),
            (defined $INFO->{last} ? strftime("%Y-%m-%d %H:%M",localtime($INFO->{last})) : 'none'),
            (defined $INFO->{count} ? $INFO->{count} : 0))
        if $verbose;
   $SENSOR =  shift @ARGV;
}
